#  Drakkar-Software OctoBot-Channels
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

"""
Handles strategy evaluations
"""
from asyncio import CancelledError

from octobot_channels import CONSUMER_CALLBACK_TYPE
from octobot_channels.channels.exchange.exchange_channel import ExchangeChannel
from octobot_channels.consumer import Consumer
from octobot_channels.producer import Producer


class TradingModeProducer(Producer):
    async def push(self, strategy_evaluation):
        # TODO replace queue content
        await self.perform(strategy_evaluation)

    async def perform(self, strategy_evaluation):
        try:
            await self.send(strategy_evaluation)
        except CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.error(f"exception when triggering update: {e}")
            self.logger.exception(e)

    async def send(self, strategy_evaluation):
        if self.channel.consumer:
            await self.channel.consumer.queue.put({"stgy_eval": strategy_evaluation})


class TradingModeConsumer(Consumer):
    async def consume(self):
        while not self.should_stop:
            try:
                data = await self.queue.get()
                await self.callback(strategy_evaluation=data["stgy_eval"])
            except Exception as e:
                self.logger.exception(f"Exception when calling callback : {e}")


class TradingModeChannel(ExchangeChannel):
    def new_consumer(self, callback: CONSUMER_CALLBACK_TYPE, size: int = 1):
        self._add_new_consumer_and_run(TradingModeConsumer(callback, size=size))



# cython: language_level=3
#  Drakkar-Software OctoBot-Channels
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
from queue import Queue

from octobot_commons.logging.logging_util import get_logger

from octobot_channels import CONSUMER_CALLBACK_TYPE


class Consumer:
    def __init__(self, callback: CONSUMER_CALLBACK_TYPE, size: int = 0, filter_size: bool = False):
        self.logger = get_logger(self.__class__.__name__)

        self.queue = Queue(maxsize=size)
        self.callback = callback
        self.consume_task = None
        self.should_stop = False
        self.filter_size = filter_size

    async def consume(self):
        """
        Should implement self.queue.get() in a while loop

        while not self.should_stop:
            self.callback(await self.queue.get())

        :return:
        """
        raise NotImplemented("consume is not implemented")

    def start(self):
        """
        Should be implemented for consumer's non-triggered tasks
        :return:
        """
        pass

    def stop(self):
        """
        Stops non-triggered tasks management
        :return:
        """
        self.should_stop = True

    def create_task(self):
        self.consume_task = asyncio.create_task(self.consume())

    def run(self):
        self.start()
        self.create_task()



# cython: language_level=3
#  Drakkar-Software OctoBot-Channels
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.logging.logging_util import get_logger
from typing import List

from octobot_channels import CONSUMER_CALLBACK_TYPE
from octobot_channels.channels.channel_instances import ChannelInstances

"""
A Channel ****
"""


class Channel(object):
    def __init__(self):
        self.producer = None
        self.consumers = {}
        self.logger = get_logger(self.__class__.__name__)

    @classmethod
    def get_name(cls) -> str:
        return cls.__name__.replace('Channel', '')

    def new_consumer(self, callback: CONSUMER_CALLBACK_TYPE, size=0, **kwargs):
        """
        Create an appropriate consumer instance for this channel
        :param callback: method that should be called when consuming the queue
        :param size: queue size, default 0
        :return: a new consumer instance
        """
        raise NotImplemented("new consumer is not implemented")

    def get_consumers(self, **kwargs) -> List:
        """
        Should be overwritten according to the class needs
        :param kwargs:
        :return:
        """
        return self.consumers

    async def start(self):
        for consumer in [consumer.values() for consumer in self.consumers.values()]:
            await consumer.start()

        if self.producer:
            await self.producer.start()

    async def stop(self):
        for consumer in [consumer.values() for consumer in self.consumers.values()]:
            await consumer.stop()

        if self.producer:
            await self.producer.stop()

    async def run(self):
        for consumer in [consumer.values() for consumer in self.consumers.values()]:
            await consumer.run()

        if self.producer:
            await self.producer.run()


class Channels:
    @staticmethod
    def set_chan(chan: Channel, name: str):
        chan_name = chan.get_name() if name else name
        if chan_name not in ChannelInstances.instance().channels:
            ChannelInstances.instance().channels[chan_name] = chan
        else:
            raise ValueError(f"Channel {chan_name} already exists.")

    @staticmethod
    def get_chan(chan_name: str, **kwargs) -> Channel:
        return ChannelInstances.instance().channels[chan_name]
